/*
 * cement.c  adds uchar image into total.pdm (double) file
 * steve; from ndf's diff.c, with help of sbeck on parsing input images
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <assert.h>
#include "cement.h"
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "pnmutils.h"


/* Global Variables */
int verbose=0;
double powLookup[256];

extern int errno;
extern char **environ;

/* Function prototypes */
int my_system (char *command);
inline double createValue(double a, unsigned char b, double weight);
void initilizeLookup(double p);
void usage(char * string);
void parse_commandline(int argc, char ** argv,
                       struct image_params * a_params,
                       struct image_params * b_params,
                       double *R,
                       double *G,
                       double *B);

int main(int argc, char ** argv)
{
    struct image_params a_params={NULL,0,0,0,0,0.0};
    struct image_params b_params={NULL,0,0,0,0,0.0};
    FILE * a, * b, * dest;
    double p=EXPONENT;  /* exponent for pow */
    double p_inv;  /* 1.0/p */
    double        a_pixel[ARRAY_SIZE];
    unsigned char b_pixel[ARRAY_SIZE];
    double out_pixel[ARRAY_SIZE];
    int i,j;
    int channels_per_pixel;
    double R, G, B;  /* weights for cement */

    char string[MAX_FILE_NAME_LENGTH];
    
    /* Timing measurments */
    struct timeval timeValStart;
    struct timeval timeValEnd;
    struct timezone timeZone;
    int minutes;
    double seconds=0.0;
#ifdef GET_TIME_MEASUREMENTS
    struct timeval runTimeStart;
    struct timeval runTimeStop;
    double fileReadTime=0.0;
    double fileWriteTime=0.0;
    double computeTime=0.0;
#endif
    
    unsigned int pixels=0;
    unsigned int numberOfPixels;
    
    p_inv=1.0/p;

    parse_commandline(argc, argv, &a_params, &b_params,
                      &R, &G, &B);

    if ((a=fopen(a_params.filename, "r"))==NULL)
    {
        fprintf(stderr, "Unable to open %s.\n", a_params.filename);
        exit(EXIT_FAILURE);
    }
    
    if ((b=fopen(b_params.filename, "r"))==NULL)
    {
        fprintf(stderr, "Unable to open %s.\n", b_params.filename);
        exit(EXIT_FAILURE);
    }

    sprintf(string,"%s.tmp",a_params.filename);
    if ((dest=fopen(string, "w"))==NULL)
    {
        fprintf(stderr, "Unable to open %s.\n", string);
        exit(EXIT_FAILURE);
    }

    get_image_type(a, &a_params);

    if (a_params.type=='A')
    {
        channels_per_pixel=3;
        fprintf(dest, "P%c\n", a_params.type);
        get_image_params_preserve_comments(a,&a_params,dest);
    }
    else
    {
        fprintf(stderr, "Portable Lightspace Map (plm) must be type A and it is type %c\n",a_params.type);
        exit(EXIT_FAILURE);
    }

    get_image_params(b, &b_params);

    
    if ((a_params.width!=b_params.width) || (a_params.height!=b_params.height)
        || (a_params.max_val!=b_params.max_val) || (a_params.exponent!=(float)EXPONENT))
    {
        printf("%f %f\n",a_params.exponent,EXPONENT);
        fprintf(stderr, "Lightspace must have the same dimensions, maximum "
                "values, and exponent.\n");
        exit(EXIT_FAILURE);
    }

    fprintf(dest, "# %s %s %s %s %s %s\n",argv[0],argv[1],argv[2],argv[3],
            argv[4],argv[5]);
    
#ifndef CEMENT_REMOVE
    fprintf(dest, "%d %d\n%d\n%d\n%f\n",a_params.width,
            a_params.height,a_params.max_val, ++a_params.numberOfImages,EXPONENT);
#else
    fprintf(dest, "%d %d\n%d\n%d\n%f\n",a_params.width,
            a_params.height, a_params.max_val,--a_params.numberOfImages,EXPONENT);
#endif

    
#ifndef CEMENT_REMOVE
    printf("Adding file: \"%s\", ",b_params.filename);
#else
    printf("Removing file: \"%s\", ",b_params.filename);
#endif
    printf("P%c, %dx%d, ",b_params.type,b_params.width,b_params.height);

    printf("R=%7.3f ",R);
    printf("G=%7.3f ",G);
    printf("B=%7.3f to ",B);

    printf("lightspace \"%s\", ",a_params.filename);
    printf("P%c\n",a_params.type);

    initilizeLookup(p);

    /* Start Timer */
    if(gettimeofday(&timeValStart,&timeZone)==-1)
    {
        fprintf(stderr,"Error getting the time\n");
        exit(0);
    }

    numberOfPixels=a_params.width*a_params.height;

    if(((float)numberOfPixels/NUMBER_PIXELS_PER_READ)!=(int)((float)numberOfPixels/NUMBER_PIXELS_PER_READ))
    {
        if(verbose)
            printf("\nERROR: The number of pixels %d is not evenly divisible by %d. ",numberOfPixels,NUMBER_PIXELS_PER_READ);

        numberOfPixels=(int)((float)numberOfPixels/NUMBER_PIXELS_PER_READ)*NUMBER_PIXELS_PER_READ;
        if(verbose)
            printf("We will go up to %d pixels.\n",numberOfPixels);
    }

    for(i=0;i<numberOfPixels;i+=NUMBER_PIXELS_PER_READ)
    {
        if(i%(3*NUMBER_PIXELS_PER_READ)==0)
        {
            printf("  %5.2f%% done.\r",(double)100*i/numberOfPixels);
            fflush(stdout);
        }

        
#ifdef GET_TIME_MEASUREMENTS
        gettimeofday(&runTimeStart,&timeZone);
#endif

        fread(a_pixel, sizeof(double),        ARRAY_SIZE, a);
        fread(b_pixel, sizeof(unsigned char), ARRAY_SIZE, b);

#ifdef GET_TIME_MEASUREMENTS
        gettimeofday(&runTimeStop,&timeZone);

        seconds=(double)runTimeStop.tv_sec+(double)runTimeStop.tv_usec/1000000;
        seconds-=(double)runTimeStart.tv_sec+(double)runTimeStart.tv_usec/1000000;
        fileReadTime+=seconds;

        gettimeofday(&runTimeStart,&timeZone);
#endif
        
        for(j=0;j<ARRAY_SIZE;j+=3)
        {
            out_pixel[j  ]=createValue(a_pixel[j  ],b_pixel[j  ],R);
            out_pixel[j+1]=createValue(a_pixel[j+1],b_pixel[j+1],G);
            out_pixel[j+2]=createValue(a_pixel[j+2],b_pixel[j+2],B);
        }

#ifdef GET_TIME_MEASUREMENTS
        gettimeofday(&runTimeStop,&timeZone);

        seconds=(double)runTimeStop.tv_sec+(double)runTimeStop.tv_usec/1000000;
        seconds-=(double)runTimeStart.tv_sec+(double)runTimeStart.tv_usec/1000000;

        computeTime+=seconds;

        gettimeofday(&runTimeStart,&timeZone);
#endif
        
        fwrite(out_pixel, sizeof(double), ARRAY_SIZE, dest);

#ifdef GET_TIME_MEASUREMENTS
        gettimeofday(&runTimeStop,&timeZone);

        seconds=(double)runTimeStop.tv_sec+(double)runTimeStop.tv_usec/1000000;
        seconds-=(double)runTimeStart.tv_sec+(double)runTimeStart.tv_usec/1000000;
        fileWriteTime+=seconds;
#endif
    }

    if(numberOfPixels!=a_params.width*a_params.height)
    {
        if(verbose)
            printf("Cleanup on pixels %d to %d\n",numberOfPixels,a_params.width*a_params.height);

        for(i=numberOfPixels;i<a_params.width*a_params.height;i++)
        {

            fread(a_pixel, sizeof(double),CHANNELS_PER_PIXEL, a);
            fread(b_pixel, sizeof(unsigned char), CHANNELS_PER_PIXEL, b);

            out_pixel[0]=createValue(a_pixel[0],b_pixel[0],R);
            out_pixel[1]=createValue(a_pixel[1],b_pixel[1],G);
            out_pixel[2]=createValue(a_pixel[2],b_pixel[2],B);
            fwrite(out_pixel, sizeof(double), CHANNELS_PER_PIXEL, dest);
            pixels++;

        }
    }

    /* End timer */
    if(gettimeofday(&timeValEnd,&timeZone)==-1)
    {
        printf("Error getting the time\n");
        exit(0);
    }

    pixels=i;

    while (!feof(a))
    {
        fread(a_pixel, sizeof(double),CHANNELS_PER_PIXEL, a);
        fread(b_pixel, sizeof(unsigned char), CHANNELS_PER_PIXEL, b);

        if(feof(a))
            break;
        
        out_pixel[0]=createValue(a_pixel[0],b_pixel[0],R);
        out_pixel[1]=createValue(a_pixel[1],b_pixel[1],G);
        out_pixel[2]=createValue(a_pixel[2],b_pixel[2],B);
        printf("EP A: %f %f %f\t",a_pixel[0],a_pixel[1],a_pixel[2]);
        printf("B: %d %d %d\t",b_pixel[0],b_pixel[1],b_pixel[2]);
        printf("OUT: %f %f %f\n",out_pixel[0],out_pixel[1],out_pixel[2]);
        fwrite(out_pixel, sizeof(double), CHANNELS_PER_PIXEL, dest);
        pixels++;
    }
    
    fclose(a);
    fclose(b);
    fclose(dest);

    seconds=(double)timeValEnd.tv_sec+(double)timeValEnd.tv_usec/1000000;
    seconds-=(double)timeValStart.tv_sec+(double)timeValStart.tv_usec/1000000;

#ifdef GET_TIME_MEASUREMENTS
    printf("%f s doing File Read %f%%\n",fileReadTime,100*fileReadTime/seconds);
    printf("%f s doing File Writes %f%%\n",fileWriteTime,100*fileWriteTime/seconds);
    printf("%f s doing Computation %f%%\n",computeTime,100*computeTime/seconds);
    printf("%f s doing Timing and looping\n",seconds-fileReadTime-fileWriteTime-computeTime);
#endif
    
    if(seconds>60)
    {
        minutes=seconds/60;
        seconds=seconds-minutes*60;
    }
    else
        minutes=0;

    if(verbose || GET_RUN_TIME)
    {
        if(minutes==1)
            printf("Run Time is about %d minute %f seconds.\n",minutes,seconds);
        else
            printf("Run Time is about %d minutes %f seconds.\n",minutes,seconds);
    }

    sprintf(string,"rm %s; mv %s.tmp  %s",a_params.filename,a_params.filename,a_params.filename);
    if(verbose)
        printf("%s\n",string);

    if(my_system (string)==-1)
    {
        printf("Error calling %s\n",string);
        exit(EXIT_FAILURE);
    }

    exit(EXIT_SUCCESS);
}

void initilizeLookup(double p)
{
    int i;

    for(i=0;i<256;i++)
        powLookup[i]=pow((double)i,p);
}

inline double createValue(double a, unsigned char b, double weight)
{
#ifndef CEMENT_REMOVE
    return(a+weight*powLookup[b]);
#else
    return(a-weight*powLookup[b]);
#endif
}

int my_system (char *command)
{
    int pid, status;

    if (command == 0)
        return 1;
    pid = fork();
    if (pid == -1)
        return -1;
    if (pid == 0)
    {
        char *argv[4];
        argv[0] = "sh";
        argv[1] = "-c";
        argv[2] = command;
        argv[3] = 0;
        execve("/bin/sh", argv, environ);
        exit(127);
    }
    do {
        if (waitpid(pid, &status, 0) == -1) {
            if (errno != EINTR)
                return -1;
        } else
            return status;
    } while(1);
}

void usage(char * string)
{
    fprintf(stderr, "Use: %s [Lightspace file] [Imagespace file] <R> <G> <B>\n",string);
    fprintf(stderr, " [Lightspace file] is Portable Lightspace Map (plm) file image to be added to\n");
    fprintf(stderr, " [Imagespace file] is ppm file of image to add to the lightspace.\n");
    fprintf(stderr, " <R>=Float for Red weight.\n");
    fprintf(stderr, " <G>=Float for Green weight.\n");
    fprintf(stderr, " <B>=Float for Blue weight.\n");
    exit(EXIT_SUCCESS);
}

void parse_commandline(int argc, char ** argv,
                       struct image_params * a_params,
                       struct image_params * b_params,
                       double *R,
                       double *G,
                       double *B)
{
    int i;

    if (argc<3)
        usage(argv[0]);

    for (i=1;i<argc;i++)
    {
        if (a_params->filename==NULL)
            a_params->filename=(char *)strdup(argv[i]); /* if file a still null */
        else if (b_params->filename==NULL)
            b_params->filename=(char *)strdup(argv[i]);
        /* else use(); */
        //*R: scanf(...R)    double must be a pointer, choice exists only with interger
        sscanf(argv[3],"%lf",R); /* sscanf does not recognize %g, must be %lf */
        //R:  sscanf(argv[3],"%lf",&R);
        sscanf(argv[4],"%lf",G); /* sscanf does not recognize %g, must be %lf */
        sscanf(argv[5],"%lf",B); /* sscanf does not recognize %g, must be %lf */
    }

    if ((a_params->filename==NULL) || (b_params->filename==NULL))
        usage(argv[0]);
}

